open Prefs
open Debug
open Tptp
open Terms
open Print
open Status
open Learning
open Unix


module type External_checker = 
sig
  val name : string
  val default_cmd : string
  val get_handle : string -> in_channel * out_channel
  val get_handles : int -> string -> (in_channel * out_channel) list
  val tell : in_channel * out_channel -> string list -> unit
  val ask : in_channel * out_channel -> checker_status
  val shutdown : unit -> unit
end
;;  

(* FIXME : I have no idea how modules work *)

module AProVE = ( Aprove : External_checker );;
module TPA = ( Tpa : External_checker );;



type checker_t = APROVE | TPA;;

let string_of_checker = function
    APROVE -> AProVE.name
  | TPA -> TPA.name
;;

let checker = ref APROVE;;
let checker_cmd = ref "aprove";;

let use_aprove() = checker_cmd := AProVE.default_cmd; checker := APROVE;;
let use_tpa() = checker_cmd := TPA.default_cmd; checker := TPA;;

let checker_name() = string_of_checker !checker;;

let get_handle cmd =
  match !checker with 
      APROVE -> AProVE.get_handle cmd
    | TPA -> TPA.get_handle cmd
;;

let get_handles n cmd = 
  match !checker with 
      APROVE -> AProVE.get_handles n cmd
    | TPA -> TPA.get_handles n cmd
;;

let tell handle msg = 
  match !checker with 
      APROVE -> AProVE.tell handle msg
    | TPA -> TPA.tell handle msg
;;

let ask handle = 
  match !checker with 
      APROVE -> AProVE.ask handle
    | TPA -> TPA.ask handle
;;

let shutdown () = 
  match !checker with 
      APROVE -> AProVE.shutdown ()
    | TPA -> TPA.shutdown ()
;;

let call_count = ref 0;;
let call_time = ref 0.;;
let max_call_time = ref 0.;;
let min_call_time = ref max_float;;
let total_yes_time = ref 0.;;
let max_yes_time = ref 0.;;
let yes_count = ref 0;;
let no_count = ref 0;;
let maybe_count = ref 0;;
let orientations = ref 0;;
let duplicates = ref 0;;

let tell_checker msgs handle = 
  incr call_count;
  debug_if_l TRACE (!debug_checker) 
    (lazy ("tell: " ^ (string_of_int !call_count)));
  tell handle msgs
;;

let ask_checker handle = 
  let start_t = gettimeofday() in
  let status = ask handle in
  let stop_t = gettimeofday() in
  let this_call_time = stop_t -. start_t in
    call_time := !call_time +. this_call_time;
    if (compare this_call_time !max_call_time) > 0 then 
      max_call_time := this_call_time;
    if (compare this_call_time !min_call_time) < 0 then 
      min_call_time := this_call_time;
    debug_if_l TRACE (!debug_checker) 
      (lazy ("ask: " ^ (string_of_int !call_count) ^ " " ^
               (string_of_checker_status status) ^ " " ^
               (string_of_float this_call_time)));
    log
      (lazy ((string_of_checker_status status) ^ " " ^
               (string_of_float this_call_time) ^ "\n"));
    begin
      match status with
          TERM_YES -> 
            incr yes_count; 
            total_yes_time := !total_yes_time +. this_call_time;
            if this_call_time > !max_yes_time then 
              max_yes_time := this_call_time;
(*             register_terminating_trs all_rules; *)
            true
        | TERM_NO -> 
            incr no_count;
(*             register_nonterminating_trs all_rules; *)
            false
        | TERM_UNKNOWN -> 
            incr maybe_count;
            false
    end,
  this_call_time
;;

let call_checker msgs handle = 
  tell_checker msgs handle;
  ask_checker handle
;;

(* let call_checker_save msgs handle =    *)
(*   incr call_count; *)
(*   debug_if_l TRACE (!debug_checker)  *)
(*     (lazy ("tell: " ^ (string_of_int !call_count))); *)
(*   tell handle msgs; *)
(*   let start_t = gettimeofday() in *)
(*   let status = ask handle in *)
(*   let stop_t = gettimeofday() in *)
(*   let this_call_time = stop_t -. start_t in *)
(*     call_time := !call_time +. this_call_time; *)
(*     if (compare this_call_time !max_call_time) > 0 then  *)
(*       max_call_time := this_call_time; *)
(*     if (compare this_call_time !min_call_time) < 0 then  *)
(*       min_call_time := this_call_time; *)
(*     debug_if_l TRACE (!debug_checker)  *)
(*       (lazy ("ask: " ^ (string_of_int !call_count) ^ " " ^ *)
(*                (string_of_checker_status status) ^ " " ^ *)
(*                (string_of_float this_call_time))); *)
(*     begin *)
(*       match status with *)
(*           TERM_YES ->  *)
(*             incr yes_count;  *)
(*             total_yes_time := !total_yes_time +. this_call_time; *)
(*             if this_call_time > !max_yes_time then  *)
(*               max_yes_time := this_call_time; *)
(* (\*             register_terminating_trs all_rules; *\) *)
(*             true *)
(*         | TERM_NO ->  *)
(*             incr no_count; *)
(* (\*             register_nonterminating_trs all_rules; *\) *)
(*             false *)
(*         | TERM_UNKNOWN ->  *)
(*             incr maybe_count; *)
(*             false *)
(*     end, *)
(*   this_call_time *)
(* ;; *)

let are_terminating rvs = 
  let rec aret acc = function 
      [] -> acc
    | rvs -> 
        let handles_wanted = 
          if !Prefs.parallel_checker then 
            (List.length rvs)
          else
            1
        in
        let handles = get_handles handles_wanted !checker_cmd in
        let probs,deferred = 
          let rec first = function
              0,xs -> []
            | n,[] -> []
            | n,x::xs -> x::(first ((n-1),xs))
          in
          let rec skip = function
              0,xs -> xs
            | n,[] -> []
            | n,x::xs -> skip ((n-1),xs)
          in
          let handle_count = List.length handles in
            first (handle_count,rvs), skip (handle_count,rvs)
        in
        let phs = List.combine probs handles in
        let acc' = 
          ((List.iter (fun ((r,v),h) -> 
                        let vars_msg = string_of_vars_tpdb v in
                        let rules_msg = string_of_trs_tpdb r in
                          tell_checker [vars_msg;rules_msg] h)) phs);
          acc@
            ((List.map (fun (p,h) -> ask_checker h)) phs)
        in
          aret acc' deferred
  in
    aret [] rvs
;;
    
let is_terminating r v =
  match are_terminating [(r,v)] with
      [true,time] -> true,time
    | xs -> false,0.0
;;

(* let is_terminating_save all_rules vars = *)
(*   match all_rules with  *)
(*       r::rs ->  *)
(*         let vars_msg = string_of_vars_tpdb vars in *)
(*         let rules_msg = string_of_trs_tpdb all_rules in *)
(*         let handle = get_handle !checker_cmd in *)
(*           incr call_count; *)
(*           debug_if_l TRACE (!debug_checker)  *)
(*             (lazy ("tell: " ^ (string_of_int !call_count))); *)
(*           tell handle [vars_msg;rules_msg]; *)
(*           let start_t = gettimeofday() in *)
(*           let status = ask handle in *)
(*           let stop_t = gettimeofday() in *)
(*           let this_call_time = stop_t -. start_t in *)
(*             call_time := !call_time +. this_call_time; *)
(*             if (compare this_call_time !max_call_time) > 0 then  *)
(*               max_call_time := this_call_time; *)
(*             if (compare this_call_time !min_call_time) < 0 then  *)
(*               min_call_time := this_call_time; *)
(*             debug_if_l TRACE (!debug_checker)  *)
(*               (lazy ("ask: " ^ (string_of_int !call_count) ^ " " ^ *)
(*                        (string_of_checker_status status) ^ " " ^ *)
(*                        (string_of_float this_call_time))); *)
(*             begin *)
(*               this_call_time, *)
(*               match status with *)
(*                   TERM_YES ->  *)
(*                     incr yes_count;  *)
(*                     total_yes_time := !total_yes_time +. this_call_time; *)
(*                     if this_call_time > !max_yes_time then  *)
(*                       max_yes_time := this_call_time; *)
(*                     register_terminating_trs all_rules; *)
(*                     true *)
(*                 | TERM_NO ->  *)
(*                     incr no_count; *)
(*                     register_nonterminating_trs all_rules; *)
(*                     false *)
(*                 | TERM_UNKNOWN ->  *)
(*                     incr maybe_count; *)
(*                     false *)
(*             end *)
(*     | [] -> failwith "Testing empty TRS for termination" *)
(* ;; *)


(* always shutdown checker on exit *)
at_exit 
  (fun () -> 
     debug (lazy ("Stopping " ^ (checker_name()))); 
     shutdown ())
;;

